###############################################################
# pytest -v --capture=no tests/test_server_cms_cpu.py
# pytest -v  tests/test_server_cms_cpu.py
# pytest -v --capture=no  tests/test_server_cms_cpu..py::Test_server_sampleFunction::<METHODNAME>
###############################################################
import pytest
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.Shell import Shell
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.util import HEADING

Benchmark.debug()

cloud = "local"

name = "sampleFunction"

@pytest.mark.incremental
class TestServerSampleFunction:

    def test_start(self):
        HEADING()

        Benchmark.Start()
        result = Shell.execute(f"cms openapi server start ./tests/{name}.yaml",
                               shell=True)
        Benchmark.Stop()
        VERBOSE(result)

        #assert False # find test

    def test_stop(self):
        HEADING()

        Benchmark.Start()
        result = Shell.execute(f"cms openapi server stop {name}", shell=True)
        Benchmark.Stop()
        VERBOSE(result)

        #assert False # find test


    def test_benchmark(self):
        HEADING()
        Benchmark.print(csv=True, tag=cloud)




from sklearn.linear_model import LinearRegression
import numpy as np
import pandas as pd
import array
from cloudmesh.openapi.registry.cache import ResultCache
from cloudmesh.openapi.registry.fileoperation import FileOperation


def fit(X: str, y: str):

    """
    Fit linear model.


    :param X: Training data
    :type X: array
    :param y: Target values. Will be cast to X's dtype if necessary
    :type y: array
    :param sample_weight: Individual weights for each sample
                    
                    .. versionadded:: 0.17
                       parameter *sample_weight* support to LinearRegression.
    :type sample_weight: array
    :param return: 
    :type return: self
    
    """

    X_input = "~/.cloudmesh/upload-file/" + f"{X}" + ".csv"
    y_input = "~/.cloudmesh/upload-file/" + f"{y}" + ".csv"
    X = pd.read_csv(X_input)
    y = pd.read_csv(y_input)
    fit = LinearRegression().fit(X, y)
    ResultCache().save("Linregnew","pickle",fit)


    return 


def get_params(deep: bool) -> str:

    """
    Get parameters for this estimator.


    :param deep: If True, will return the parameters for this estimator and
                    contained subobjects that are estimators.
    :type deep: bool
    :param return: Parameter names mapped to their values.
    :type return: str
    
    """

    model = ResultCache().load("Linregnew")
    str = model.get_params(deep)


    return str


def predict(X: str) -> list:

    """
    Predict using the linear model.


    :param X: Samples.
    :type X: array
    :param return: Returns predicted values.
    :type return: array
    
    """

    X_input = "~/.cloudmesh/upload-file/" + f"{X}" + ".csv"
    X = pd.read_csv(X_input)
    model = ResultCache().load("Linregnew")
    list = model.predict(X)
    list = list.tolist()


    return list


def score(X: str, y: str) -> float:

    """
    Return the coefficient of determination R^2 of the prediction.


    :param X: Test samples. For some estimators this may be a
                    precomputed kernel matrix or a list of generic objects instead,
                    shape = (n_samples, n_samples_fitted),
                    where n_samples_fitted is the number of
                    samples used in the fitting for the estimator.
    :type X: array
    :param y: True values for X.
    :type y: array
    :param sample_weight: Sample weights.
    :type sample_weight: array
    :param return: R^2 of self.predict(X) wrt. y.
    :type return: float
    
    """

    X_input = "~/.cloudmesh/upload-file/" + f"{X}" + ".csv"
    y_input = "~/.cloudmesh/upload-file/" + f"{y}" + ".csv"
    X = pd.read_csv(X_input)
    y = pd.read_csv(y_input)
    model = ResultCache().load("Linregnew")
    float = model.score(X, y)


    return float


def set_params(**params: dict):

    """
    Set the parameters of this estimator.


    :param **params: Estimator parameters.
    :type **params: dict
    :param return: Estimator instance.
    :type return: self
    
    """

    set_params = LinearRegression().set_params(**params)
    ResultCache().save("Linregnew","pickle",set_params)


    return



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
import os
import time
from pprint import pprint

import pytest
import tests.util as util
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import HEADING
from cloudmesh.common.variable import Variables

variable = Variables()
filename = variable['filename']


# filename = "./tests/server-cpu/cpu.yaml"

@pytest.mark.incremental
class TestGenerator:

    def test_read_spec(self):
        """
        function to check if YAML synatx is correct or not
        """
        global spec

        HEADING()
        Benchmark.Start()
        spec = util.readyaml(filename)
        keys = spec.keys()

        pprint(spec)
        pprint(keys)
        assert "openapi" in keys
        assert "info" in keys
        assert "servers" in keys
        assert "paths" in keys
        assert "A simple service" in str(spec)

    def test_paths(self):
        """
        function to validate paths information
        """
        HEADING()
        global spec

        import tests.sample_function_gen as testfun

        paths = spec.get("paths")
        name = testfun.sampleFunction.__name__
        getOperation = paths.get(f"/{name}")

        assert paths is not None, "paths value should not be null"
        assert paths.keys().__contains__(f"/{name}"), "Resource name should be {name}"
        assert getOperation.keys().__contains__("get"), "get operation is missing "
        parameters = getOperation.get("get").get("parameters")
        # assert len(parameters)+1==len(testfun.sampleFunction.__annotations__.items()), "get operation is missing "

        Benchmark.Stop()

    def test_service_start(self):
        """
        function to test if the server is starts
        """
        HEADING()

        os.chdir("tests")
        test_loc = Shell.pwd().strip() + "/"

        assert test_loc == "tests/"

        Benchmark.Start()
        server_output = Shell.cms("openapi server start {filename}")
        Benchmark.Start()

        assert server_output.__contains__("starting server")

    def test_service_get(self):
        """
        function to test if the server is returns the result from the url
        """
        HEADING()

        time.sleep(2)
        baseurl = "http://127.0.0.1:8080/cloudmesh"

        curl = f"curl {baseurl}/cpu"

        Benchmark.Start()
        response = Shell.run(curl)
        Benchmark.Stop()

        assert response.__contains__("200")

    def test_service_stop(self):
        """
        function to test if the server is returns the result from the url
        """
        HEADING()

        Shell.cms("openapi server stop cpu")
        response = Shell.run(curl)
        fail_message = "Failed to connect to 127.0.0.1 port 80: Connection refuse"
        assert response.__contains__(fail_message)

        Benchmark.Stop()

    def test_benchmark(self):
        HEADING()
        Benchmark.print(csv=True, sysinfo=False, tag="generator")



# noinspection PyPep8Naming
def sampleFunction(x: int, y: float) -> float:
    """
    Multiply int and float sample.

    :param x: x value
    :type x: int
    :param y: y value
    :type y: float
    :return: result
    :return type: float
    """
    return x * y

def sampleFunction2(x2: int, y2: float) -> float:
    """
    Multiply int and float sample.

    :param x2: x value
    :type x2: int
    :param y2: y value
    :type y2: float
    :return: result
    :return type: float
    """
    return x2 * y2




import inspect
import re
import textwrap
from pydoc import locate
from numpydoc import docscrape


class TypeScraper:
    """
    Scrape types from a string.
    Using  the regular expression to match the keywords that imply the
    types.

    A type table for matching the types from the string is required

    Examples:
         'boolean, optional, default True' = bool
         'int or None, optional (default=None)' = int
         'array-like or sparse matrix, shape (n_samples, n_features)' = list
         'numpy array of shape [n_samples]' 'boolean, optional' = list
    """

    def __init__(self, type_table):
        """
        The Constructor function

            Parameters:
                type_table: A dictionary indicates the matching rules
        """
        self.type_table = type_table

    def scrap(self, literal_type):
        """
        Match types from the string

            Parameters:
                literal_type: A string that defines a type
        """

        res = set()

        # Traverse all known mappings to check which key of the table
        # matches the string
        for table_key in self.type_table.keys():
            #print(literal_type)
            if re.search(table_key, literal_type, re.IGNORECASE):
                res.add(self.type_table[table_key])

        #if literal_type[0] == '{':
        #    table_key = 'string'
        #    res.add(self.type_table[table_key])
        if literal_type[:4] == 'dict':
            table_key = 'dictionary'
            res.add(self.type_table[table_key])
        if len(res) == 1:
            return res.pop()
        else:
            reslist = list(res)
            return reslist[0]


class Generator:

    importfuction = textwrap.dedent("""
        from {library}.{module} import {class_nm}
        import numpy as np
        import array
        from cloudmesh.openapi.registry.cache import ResultCache
                """)

    functiontemplatearrayandX_param = textwrap.dedent("""
    
        def {functioname}({parameters}) -> {returnparam1}:
        
            {text1}
            {description}
            
            
            {docstring}
            {text1}
            
            {X_numpyconversion}
            model = ResultCache().load("{model_tag}")
            {returnparam1} = model.{functioname}({param_wo_type})
            {returnparam1} = {returnparam1}.tolist()
            
            
            return {returnparam1}
        """)

    functiontemplatewithX_param = textwrap.dedent("""

            def {functioname}({parameters}) -> {returnparam1}:

                {text1}
                {description}


                {docstring}
                {text1}

                {X_numpyconversion}
                model = ResultCache().load("{model_tag}")
                {returnparam1} = model.{functioname}({param_wo_type})
                
                
                return {returnparam1}
            """)

    functiontemplate = textwrap.dedent("""

        def {functioname}({parameters}) -> {returnparam1}:

            {text1}
            {description}


            {docstring}
            {text1}

            model = ResultCache().load("{model_tag}")
            {returnparam1} = model.{functioname}({param_wo_type})
            
            
            return {returnparam1}
        """)

    functiontemplatereturningself = textwrap.dedent("""
    
            def {functioname}({parameters}):

                {text1}
                {description}


                {docstring}
                {text1}
               
                {functioname} = {base_estimator}().{functioname}({param_wo_type})
                ResultCache().save("{model_tag}","pickle",{functioname})
                


                return
            """)

    functiontemplatefit = textwrap.dedent("""

                def {functioname}({parameters}):

                    {text1}
                    {description}


                    {docstring}
                    {text1}
                    
                    {X_numpyconversion}
                    {functioname} = {base_estimator}().{functioname}({param_wo_type})
                    ResultCache().save("{model_tag}","pickle",{functioname})


                    return 
                """)

    functiontemplatesetparams = textwrap.dedent("""

                    def {functioname}({parameters}):

                        {text1}
                        {description}


                        {docstring}
                        {text1}

                        {functioname} = {base_estimator}().{functioname}({param_wo_type})
                        ResultCache().save("{model_tag}","pickle",{functioname})


                        return
                    """)
    def populate_parameters_function(self, f, paras_dict, paras_desc):
        """
        Function to loop all the parameters of given function and generate
        specification

        :param paras_dict: parameters dictionary passed to the function
        :param paras_desc: parameters docstrings passed to the function
        :return:
        """
        spec = str()
        spec_params = str()
        docstring = str()
        for i, item in enumerate(paras_dict):
            if i == (len(paras_dict)-1):
                last_key = item
        for parameter_dict, _type in paras_dict.items():
            for parameter_desc, desc in paras_desc.items():
                if parameter_dict == 'return1':
                    docstring = docstring + \
                                (f""":param {parameter_dict}: {desc}""") + "\n" + \
                                "    " + (f""":type {parameter_dict}: {_type}""") + "\n" + "    "
                else:
                    if parameter_dict == parameter_desc:
                        docstring = docstring  + \
                        (f""":param {parameter_dict}: {desc}""") +  "\n" + \
                        "    "+ (f""":type {parameter_dict}: {_type}""") + "\n" + "    "
                        if parameter_dict == last_key:
                            spec = spec + (f"""{parameter_dict}: {_type}""")
                            spec_params =  spec_params + (f"""{parameter_dict}""")
                        else:
                            spec = spec + f"""{parameter_dict}: {_type},""" + " "
                            spec_params = spec_params + f"""{parameter_dict},""" + " "

        return spec,spec_params,docstring

    def is_valid_para(self, para_type, type_table):
        """Check if it is a valid parameter type contained in the type table.
        """
        # The values of the table contain all known destination types
        if para_type in type_table.values():
            return True
        return True

    def get_parameters(self, parsing_obj, type_table):
        """Get parameters from the doc of a class, function, or property object.

        Given the sklean docstring follows the numpy conventions, this function
        use the numpy docstring parser to read the doc of sklean.
        """
        scraper = TypeScraper(type_table=type_table)
        r = parsing_obj
        paras = {}
        returnparam = {}
        for p in r['Parameters']:
            para_str = str(p.type)
            para_type = scraper.scrap(para_str)
            if self.is_valid_para(para_type, type_table):
                paras[p.name] = para_type
            else:
                continue

        for p in r['Returns']:
            para_str = str(p.type)
            para_type = scraper.scrap(para_str)
            if self.is_valid_para(para_type, type_table):
                #returnparam['return'] = para_type
                paras['return'] = para_type
            else:
                continue
        key = 'return'
        if key not in paras.keys():
            paras['return'] = 'self'
        return paras

    def get_docstrings(self, parsing_obj):
        """Get descriptions  from the doc of a class, function, or property object.

        Given the sklean docstring follows the numpy conventions, this function
        use the numpy docstring parser o read the doc of sklean.
        """
        #r = docscrape.NumpyDocString(doc)
        r = parsing_obj
        paras_desc = {}
        for p in r['Parameters']:
            para_name = str(p.name)
            para_desc = '\n                    '.join(p.desc)
            paras_desc[para_name] = para_desc

        for p in r['Returns']:
            #print(p.name,p.desc)
            para_name = str(p.name)
            para_desc = '\n                    '.join(p.desc)
            paras_desc['return'] = para_desc
        key = 'return'
        if key not in paras_desc.keys():
            paras_desc['return'] = 'self'
        return paras_desc

    def generate_function(self, module, function,base_estimator,model_tag):
        """

        :param module: Sklearn module like Linear Regression
        :param function: Methods in the class of Linear Regression
        :param base_estimator: Sklearn module like Linear Regression
        :param model_tag: Tag used to store the name of the model instance
        :return:
        """


        type_table = {
            'matrix': 'array',
            'array': 'array',
            'array-like': 'array',
            'numpy array': 'array',
            'bool': 'bool',
            'int': 'int',
            'float': 'float',
            'string': 'str',
            'dictionary': 'dict',
            'dict': 'dict',
            'object': 'self',
            'self': 'self'
        }
        module = module
        class_name = function
        base_estimator = base_estimator
        model_tag = model_tag
        class_obj = getattr(module, class_name)
        parsing_obj = docscrape.FunctionDoc(class_obj)
        doc = class_obj.__doc__
        paras_dict_func = self.get_parameters(parsing_obj, type_table)
        paras_desc = self.get_docstrings(parsing_obj)

        description = class_obj.__doc__.strip().split("\n")[0]
        #title = class_obj.__name__
        parametersfunc,params,docstring = self.populate_parameters_function(function, paras_dict_func, paras_desc)
        text1 = '"""'
        key = 'return'
        returnparam =''
        #print(paras_dict_func)
        if  key in paras_dict_func.keys():
            if paras_dict_func['return'] != 'self':
                returnparam = paras_dict_func['return']
            else:
                returnparam
        returnparamindex = 0
        returnparamindex = parametersfunc.find('return')
        if (returnparamindex == -1):
            pass
        elif (returnparamindex == 0):
            parametersfunc = ''
        else:
            parametersfunc = parametersfunc[:returnparamindex-2]
        returnparamindex1 = 0
        returnparamindex1 = params.find('return')
        if (returnparamindex1 == -1):
            pass
        elif (returnparamindex1 == 0):
            params = ''
        else:
            params = params[:returnparamindex1 - 2]
        functionname = class_obj.__name__
        match = re.search(r'X: array', parametersfunc)
        X_param_index = parametersfunc.find('X: array')
        sample_weight_index = parametersfunc.find('sample_weight: array')
        if sample_weight_index != -1:
            parametersfunc = parametersfunc[:sample_weight_index - 2]
        sample_weight_index_params = params.find('sample_weight')
        if sample_weight_index_params != -1:
            params = params[:sample_weight_index_params - 2]
        if match:
            parametersfunc =  parametersfunc + "," + " X_shape_x: int," + " X_shape_y: int"
        parametersfunc = parametersfunc.replace('array','list')
        X_numpyconversion = f"X = np.array(X).reshape(X_shape_x,X_shape_y)"

        if returnparam != '':
            if returnparam == 'array' and  X_param_index == 0 :
                returnparam = 'list'

                spec = self.functiontemplatearrayandX_param.format(
                    functioname=functionname,
                    description=description,
                    text1=text1,
                    model_tag=model_tag,
                    X_numpyconversion=X_numpyconversion,
                    parameters=parametersfunc,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )
                return spec
            elif X_param_index == 0 :
                spec = self.functiontemplatewithX_param.format(
                    functioname=functionname,
                    description=description,
                    text1=text1,
                    model_tag=model_tag,
                    X_numpyconversion=X_numpyconversion,
                    parameters=parametersfunc,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )
                return spec
            else:

                spec = self.functiontemplate.format(
                    functioname=functionname,
                    description=description,
                    text1 = text1,
                    model_tag=model_tag,
                    parameters=parametersfunc,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )

                return spec
        else:
            if functionname == 'fit':
                spec = self.functiontemplatefit.format(
                    functioname=functionname,
                    description=description,
                    base_estimator=base_estimator,
                    X_numpyconversion=X_numpyconversion,
                    model_tag=model_tag,
                    text1=text1,
                    parameters=parametersfunc,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )
                return spec
            elif functionname =='set_params':
                spec = self.functiontemplatesetparams.format(
                    functioname=functionname,
                    description=description,
                    base_estimator=base_estimator,
                    model_tag=model_tag,
                    text1=text1,
                    parameters=parametersfunc,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )
                return spec
            else:
                spec = self.functiontemplatereturningself.format(
                    functioname=functionname,
                    description=description,
                    text1=text1,
                    parameters=parametersfunc,
                    base_estimator=base_estimator,
                    model_tag=model_tag,
                    param_wo_type=params,
                    docstring=docstring,
                    returnparam1=returnparam
                )

                return spec

    def generate_import_params(self,input_params):
        """
        :param input_params: takes the input passed from the function
        :return:
        """

        input_params = input_params
        spec = self.importfuction.format(
                library=input_params[0],
                module=input_params[1],
                class_nm=input_params[-1]
        )
        return spec


def Sklearngenerator(input_sklibrary,model_tag):
    """

    :param input_sklibrary: input_sklibrary = sklearn.linear_model.LinearRegression(Full model specification)
    :param model_tag: model_tag = any name which you want the tag the model instance like LinReg1
    :return: .py file which is input to generator which generates openAPI specification
    """

    my_class = locate(input_sklibrary)
    method_list = [func for func,value in inspect.getmembers(my_class) if func[0] != '_']
    method_list = [value for value in method_list if value != 'classes_']
    input_params = input_sklibrary.split('.')
    input_module = f'{input_params[0]}.{input_params[1]}'
    base_estimator = input_params[-1]
    module = locate(input_module)
    class_name = input_params[-1]
    openAPI = Generator()
    spec = openAPI.generate_import_params(input_params)
    print(f"Writing python code to file: {input_params[-1]}.py")
    open(f"./tests/generator/{input_params[-1]}.py", 'w').write(spec)
    #spec = openAPI.generate_function(module, class_name,base_estimator)
    #open(f"{input_params[-1]}.py", 'a').write(spec)
    for i in range(len(method_list)):
        module = my_class
        function = method_list[i]
        openAPI = Generator()
        spec = openAPI.generate_function(module,function,base_estimator,model_tag)
        open(f"./tests/generator/{input_params[-1]}.py", 'a').write(spec)

if __name__ == "__main__":
    Sklearngenerator(input_sklibrary,model_tag)







from tests.generator import LinearRegression
import numpy as np
from sklearn.datasets import load_iris
#X, y = load_iris(return_X_y=True)
X = "X.csv"
y = "y.csv"

sample_weight = "sample_weight.csv"
#print(np.array(y))

fit = LinearRegression.fit(X,y)
print(LinearRegression.predict(X))
# print(LogisticRegression.decision_function(X,4,2))
#print(LinearRegression.predict_proba(X,4,2))
print(LinearRegression.score(X,y))

# from tests.generator import LinearRegression
# from sklearn.datasets import load_iris
# X, y = load_iris(return_X_y=True)
# print(LinearRegression.fit(X,y,None))
# print(LinearRegression.predict(X[:2, :]))
# print(LinearRegression.score(X,y,None))

# from tests.generator import RidgeClassifierCV as model
# from sklearn.datasets import load_iris
#
# X, y = load_iris(return_X_y=True)
# #print(X,y)
# print(model.fit(X,y,None))
# print(model.predict(X[:2, :]))
# print(model.score(X,y,None))




import os
import sys
import textwrap

from cloudmesh.common.console import Console
from cloudmesh.common.util import path_expand


class Parameter:
    """
    To generate a useful output for the variables. Example:

        from cloudmesh.openapi.function.executor import Parameter
        p = Parameter(arguments)
        p.Print()

    Invocation from program

        cd cloudmesh-openapi
        cms openapi generate calculator  \
            --filename=./tests/generator-calculator/calculator.py \
            --all_functions

    Returns

        Cloudmesh OpenAPI Generator:

          File Locations:
            - Currdir:    .
            - Directory:  ./tests/generator-calculator
            - Filename:   ./tests/generator-calculator/calculator.py
            - YAML:       ./tests/generator-calculator/calculator.yaml

          Yaml File Related:
            - Function:   calculator
            - Server url: http://localhost:8080/cloudmesh
            - Module:     calculator

    """

    def __init__(self, arguments):
        self.arguments = arguments
        self.filename = None
        self.module_directory = None
        self.module_name = None
        self.yamlfile = None
        self.yamldirectory = None
        self.function = None
        self.serverurl = None
        self.import_class = None
        self.all_functions = None
        self.basic_auth = None
        self.get(arguments)
        pass

    def get(self, arguments):
        self.cwd = path_expand(os.path.curdir)
        filename = arguments['--filename']
        if filename is None:
            Console.error(f"--filename={filename}")
        self.filename = path_expand(filename)
        if not os.path.isfile(filename):
            Console.error(f"--filename={self.filename} does not exist")
        
        self.module_directory = os.path.dirname(self.filename)
        self.module_name = os.path.basename(self.filename).split('.')[0]
        sys.path.append(self.module_directory)

        self.yamlfile = arguments.yamlfile or self.filename.rsplit(".py")[0] + ".yaml"
        self.yamldirectory = os.path.dirname(self.yamlfile)

        self.function = arguments.FUNCTION or os.path.basename(self.filename).split('.')[0]
        self.serverurl = arguments.serverurl or "http://localhost:8080/cloudmesh"
        self.import_class = arguments.import_class or False
        self.all_functions =arguments.all_functions or False
        self.basic_auth = arguments.basic_auth

        
    def Print(self):

        Console.info(textwrap.dedent(f"""
             Cloudmesh OpenAPI Generator:

               File Locations:
                 - Currdir:    .
                 - Filename:   {self.filename.replace(self.cwd, ".")}
                 - YAML:       {self.yamlfile.replace(self.cwd, ".")}

               Yaml File Related:
                 - Function:   {self.function}
                 - Server url: {self.serverurl}
                 - Module:     {self.module_name}

         """))



import connexion
from pathlib import Path
from cloudmesh.common.util import path_expand


class FileOperation:
    """
      This class will help to upload file from restful API
    """
    def file_upload(self) -> str:
        """
           This function will upload file into .cloudmesh/upload-file location
           This will first get upload file object from request.files function
           and then override this object in given location.
           """
        file = connexion.request.files.get("upload")
        filename = file.filename
        if file:
            file_path = f"~/.cloudmesh/upload-file"
            p = Path(path_expand(file_path))
            p.mkdir(parents=True, exist_ok=True)
            file.save(f'{p.absolute()}/{filename}')
        return filename



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
from pprint import pprint

import pytest
import yaml as yaml
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.util import HEADING
from cloudmesh.openapi.registry.Registry import Registry
from cloudmesh.common.variables import Variables
variable=Variables()
filename= variable['filename']
# sys.path.append("cloudmesh/openapi/function")
#


#
# get the spec for the tests
#

with open(filename, "r") as stream:
    try:
        spec = yaml.safe_load(stream)
    except yaml.YAMLError as e:
        print(e)
        assert False, "Yaml file has syntax error"


@pytest.mark.incremental
class TestGenerator:


    def test_registry_add(self):
        HEADING()

        Benchmark.Start()

        title = spec["info"]["title"]
        url = spec["servers"][0]["url"]

        print(f"add {title} -> {url}")
        registry = Registry()

        before = len(registry.list(name=title))

        pid = 1

        entry = registry.add(name=title, url=url, pid=pid)
        pprint (entry)

        after = len(registry.list(name=title))

        assert after == before + 1

        Benchmark.Stop()

    def test_registry_list_name(self):
        HEADING()

        Benchmark.Start()

        title = spec["info"]["title"]

        registry = Registry()

        entry = registry.list(name=title)
        pprint(entry)

        assert entry != None

        Benchmark.Stop()

    def test_registry_list(self):
        HEADING()

        Benchmark.Start()

        title = spec["info"]["title"]

        #print(f"delete {title}")
        registry = Registry()

        entry = registry.list()
        pprint(entry)

        assert len(entry) > 0

        Benchmark.Stop()

    def test_registry_delete(self):
        HEADING()


        Benchmark.Start()

        title = spec["info"]["title"]

        registry = Registry()
        entry = registry.list(name=title)

        # list before and use len()
        #before = 1
        before = len(entry)

        print(f"delete {title}")

        entry = registry.delete(name=title)
        print(f"{entry} entry deleted")

        # list after and use len

        after = len(registry.list(name=title)) # use len()

        assert before == after + 1

        Benchmark.Stop()


    def test_benchmark(self):
        HEADING()
        Benchmark.print(csv=True, sysinfo=False, tag="generator")



###############################################################
# pytest -v --capture=no tests/test_server_cms_cpu.py
# pytest -v  tests/test_server_cms_cpu.py
# pytest -v --capture=no  tests/test_server_cms_cpu..py::Test_server_cms::<METHODNAME>
###############################################################
import os
import time

import pytest
import requests
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.util import HEADING

Benchmark.debug()

cloud = "local"

name = "cpu"
yaml_file = f"./tests/server-cpu/{name}.yaml"
@pytest.mark.incremental
class TestServerCms:

    def test_start(self):
        HEADING()

        Benchmark.Start()

        os.system(f"cms openapi server start {yaml_file} --directory=./tests/server-cpu/")
        time.sleep(4)
        result = requests.get('http://127.0.0.1:8080/cloudmesh/ui')
        Benchmark.Stop()

        assert result.status_code == 200 # find test


    def test_cpu(self):
        HEADING()
        url = "http://127.0.0.1:8080/cloudmesh/cpu"
        Benchmark.Start()
        result = requests.get(url)
        time.sleep(4)
        assert result.status_code == 200
        assert result.reason == 'OK'
        assert result.headers['content-type'] =='application/json'
        assert result.json().get('model') is not None
        Benchmark.Stop()
        VERBOSE(result)

    def test_stop(self):
        HEADING()

        Benchmark.Start()
        os.system(f"cms openapi server stop {name}")
        Benchmark.Stop()
        gotException=False;
        try:
            result = requests.get('http://127.0.0.1:8080/cloudmesh/ui')
        except Exception as ex:
            gotException=True
        assert gotException



from cloudmesh.mongo.CmDatabase import CmDatabase
from cloudmesh.mongo.DataBaseDecorator import DatabaseUpdate
from pathlib import Path
from cloudmesh.common.util import path_expand
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.console import Console
import pickle
from cloudmesh.common.StopWatch import StopWatch
from cloudmesh.common.Benchmark import Benchmark
import os


class ResultCache:
    """
    Saves serialized model to local cache and saves metadata about the model to local db
    """

    def __init__(self):
        pass

    @DatabaseUpdate()
    def save(self, modelname=None, type="pickle", modelobject=None, **kwargs):
        """
        Save model to cache

        :param modelname:  the name of the model.  Will be used to name registry entry.
        :param type:  the type of serialization.  Default is pickle.
        :param modelobject:  the model object
        :param kwargs:  any other parameters for registry
        :return:  none
        """

        # create local cache directory
        cache_path = f"~/.cloudmesh/server-cache/{modelname}/"
        p = Path(path_expand(cache_path))
        p.mkdir(parents=True, exist_ok=True)

        # serialize model and save to local cache directory
        cached_file = ""
        if type == "pickle":
            cached_file = self._make_pickle(modelname, modelobject, str(p.absolute()))
        else:
            Console.error(f"Unsupported serialization type provided : {type}")
            raise Exception

        # update db cache with below details
        entry = {
            "cm": {
                "cloud": "local",
                "kind": "cache",
                "name": modelname,
                "driver": None
            },
            "name": modelname,
            "status": "cached",
            "cached_file": cached_file
        }

        for key in kwargs:
            entry[key] = kwargs[key]

        return entry

    def load(self, name):
        """
        Load cached model

        :param name:  model name or cached file name
        :return:  unserialized model
        """

        cm = CmDatabase()
        # USER env variable is required by StopWatch
        if os.getenv('USER'):
            # Do nothing
            VERBOSE("USER env variable is already defined")
        else:
            os.environ['USER'] = 'No USER env var defined'

        test = cm.find(cloud="local", kind="cache", query={"name": {'$regex': f"{name}"}})
        cached_file = test[0]['cached_file']
        Console.info(f"Loading serialized model: {cached_file}")
        StopWatch.start(f"Load pickle {name}")
        deserialized_model = self._load_pickle(cached_file)
        StopWatch.stop(f"Load pickle {name}")
        time_taken = StopWatch.get(f"Load pickle {name}")

        # TODO: figure out how useful the duration is and return to client if required
        deserialized_model_dict = {
            "model_name": name,
            "model_object": deserialized_model,
            "duration": time_taken  # duration of deserialization function
        }
        return deserialized_model

    def _make_pickle(self, title, data, path):
        """
        Serializes a model and returns the fully qualified path and file name to pickle file

        :param title:  pickle file name
        :param data:  model object
        :param path:  path to pickle file 
        :return:  pickle file name
        """

        file = Path(f"{path}/{title}.pickle")
        pikd = open(file, "wb")
        pickle.dump(data, pikd)
        pikd.close()

        return str(file)

    # loads and returns a pickled object
    def _load_pickle(self, file):
        """
        Loads a pickle file and returns object

        :param file:  pickle file name
        :return:  unpickled model object
        """

        pikd = open(file, "rb")
        data = pickle.load(pikd)
        pikd.close()
        return data


if __name__ == "__main__":

    from sklearn.linear_model import LogisticRegression
    from sklearn.datasets import load_iris


    newcache = ResultCache()

    X, y = load_iris(return_X_y=True)
    # print(X)
    clf = LogisticRegression(random_state=0, max_iter=300).fit(X, y)
    print(newcache.save("irismodel1", "pickle", clf))
    print("finished caching model")
    #model_dict = newcache.load("irismodel1")
    model = newcache.load("irismodel1")
    #print(model_dict['model_object'].predict(X[:2, :]))
    print(model.predict(X[:2, :]))



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
"""
# Headline

Here come document for test

"""
import time
from pprint import pprint
import sys
import pytest
import tests.util as util
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import HEADING
from cloudmesh.common.util import path_expand
from importlib import import_module
from cloudmesh.openapi.function.executor import Parameter
from cloudmesh.common.dotdict import dotdict
import types
from cloudmesh.shell.command import command, map_parameters

# for testing file which has function more than one, use below variable and also use dataforfunction dict
argumentsforfunction="cms openapi generate Calculator --filename=./tests/generator-calculator/calculator.py --all_functions"

# for testing file which is class , use below variable and also use dataforclass dict
argumentsforclass="cms openapi generate Calculator --filename=./tests/generator-testclass/calculator.py --import_class"

dataforclass = dotdict({
            "filename": './tests/generator-testclass/calculator.py',
            "import_class": True,
             "FUNCTION" : "Calculator",
            "all_functions":False,
        })

dataforfunction = dotdict({
            "filename": './tests/generator-testclass/calculator.py',
            "import_class": False,
            "all_functions":True,
        })

p = Parameter(dataforclass)

@pytest.mark.incremental
class TestGenerator:

    def test_generate(self):
        """
        function to validate paths information
        """
        HEADING()
        Benchmark.Start()
        Shell.run(argumentsforclass) #change variable based on your needs
        Benchmark.Stop()

    def test_read_spec(self):
        """
        function to check if YAML synatx is correct or not
        """
        global spec
        HEADING()
        Benchmark.Start()
        spec = util.readyaml(p.yamlfile)
        keys = spec.keys()
        assert "openapi" in keys
        assert "info" in keys
        assert "servers" in keys
        assert "paths" in keys

    def test_number_of_function(self):
        """
        function to check number of functions are same in py and yaml file.
        """
        HEADING()
        Benchmark.Start()
        sys.path.append(p.yamldirectory)
        imported_module = import_module(p.module_name)
        keys = spec.get('paths')
        paths= keys.keys()
        if (p.all_functions is True):
            for attr_name in dir(imported_module):
                if type(getattr(imported_module, attr_name)).__name__ == 'function':
                    assert f"/{p.function}/{attr_name}" in paths
        if (p.import_class is True):
            class_obj = getattr(imported_module, p.function)
            for attr_name in dir(class_obj):
                attr = getattr(class_obj, attr_name)
                if isinstance(attr, types.MethodType):
                    assert f"/{p.function}/{attr_name}" in paths

class rest:

    def test_service(self):
        """
        function to test if the server is started and available to return
        a successful http code
        """
        HEADING()
        Benchmark.Start()

        server_output = Shell.cms("openapi server start {filename}")
        assert server_output.__contains__("starting server")

        time.sleep(2)
        baseurl = "http://127.0.0.1:8080/cloudmesh"

        curl = f"curl {baseurl}/cpu"

        response = Shell.run(curl)
        assert response.__contains__("200")

        Shell.cms("openapi server stop cpu")
        response = Shell.run(curl)
        fail_message = "Failed to connect to 127.0.0.1 port 80: Connection refuse"
        assert response.__contains__(fail_message)

        Benchmark.Stop()




import os
import subprocess
import sys
import textwrap
from datetime import date
from importlib import import_module
from pathlib import Path

import connexion
import yaml
from cloudmesh.common.Shell import Shell
from cloudmesh.common.console import Console
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.util import path_expand
from cloudmesh.openapi.registry.Registry import Registry


def daemon(func):
    """
    Decorator used to execute a Connexion flask app as a daemon

    :param func:  function to run as daemon
    :return:  
    """
    def wrapper(*args, **kwargs):
        pid = os.fork()
        global daemonpid
        if pid != 0:
            daemonpid = pid
            print("Deamon PID", pid)

        if pid: return
        r = func(*args, **kwargs)

        os._exit(os.EX_OK)
    return wrapper


def dynamic_import(abs_module_path, class_name):
    module_object = import_module(abs_module_path)
    target_class = getattr(module_object, class_name)
    return target_class


class Server(object):
    """
    This class manages all actions taken to interact with an OpenAPI AI server.

    """

    def __init__(self,
                 name=None,
                 spec=None,
                 directory=None,
                 host="127.0.0.1",
                 server="flask",
                 port=8080,
                 debug=True):
        """
        Default constructor to initialize Server object

        :param name:  server name
        :param spec:  openapi spec yaml file
        :param directory:  directory for input file
        :param host:  host ip or dns name to be used to start service.  Default is 127.0.0.1
        :param server:  type of service to start.  Default is flask.
        :param port:  port to use for service.  Default is 8080. 
        :param debug:  flag to turn on debug logging.  Default is true.
        :return:  server object
        
        """

        if spec is None:
            Console.error("No service specification file defined")
            raise FileNotFoundError

        self.spec = path_expand(spec)

        self.name = Server.get_name(name, self.spec)

        if directory is None:
            self.directory = os.path.dirname(self.spec)
        else:
            self.directory = directory

        self.host = host or "127.0.0.1"
        self.port = port or 8080
        self.debug = debug or False
        self.server = server or "flask"
        self.server_command = ""

        data = dict(self.__dict__)
        #data['import'] = __name__

        VERBOSE(data, label="Server parameters")

        if server == "tornado":
            try:
                import tornado
            except Exception as e:
                print(e)
                Console.error(
                    "tornado not install. Please use `pip install tornado`")
                sys.exit(1)
                # return ""
            if self.debug:
                Console.error("Tornado does not support --verbose")
                sys.exit(1)
                # return ""

        Console.ok(self.directory)

    @staticmethod
    def get_name(name, spec):
        """
        Get the name of a server using specification

        :param name:  server name
        :param spec:  spec file name with fully qualified path
        :return:  server name
        """
        if name is None:
            return os.path.basename(spec).replace(".yaml", "")
        else:
            return name

    @daemon
    def _run_deamon(self):
        """
        Starts up a Connexion flask app as a daemon process

        :return:
        """
        self._run_app()

    def _run_app(self):
        """
        Starts up a Connexion flask app

        :return:
        """
        Console.ok("starting server")

        sys.path.append(self.directory)
        app = connexion.App(__name__,
                            specification_dir=self.directory)

        # ### app.add_cls(self.directory)
        app.add_api(self.spec)
        r = app.run(host=self.host,
                    port=self.port,
                    debug=self.debug,
                    server=self.server)

    def start(self, name=None, spec=None, foreground=False):
        """
        Start up an OpenApi server

        :param name:  server name
        :param spec:  openapi spec yaml file name
        :param foreground:  flag to run server in foreground.  Default is False.
        :return: started server PID
        """
        name = Server.get_name(name, spec)
        pid = ""

        if foreground:
            self._run_app()
        else:
            # We need below if condition for window machine. In window machine pid = os.fork()
            # does not work. So do not remove if condition.
            if sys.platform == 'win32':
                pid = self.run_os()
            else:
                self._run_deamon()
                if daemonpid is not None:
                    pid = daemonpid
                else:
                    pid = Server.ps(name=name)[1]["pid"]

                _spec = Server.ps(name=name)[1]["spec"]

                with open(_spec, "r") as stream:
                    try:
                        details = yaml.safe_load(stream)
                    except yaml.YAMLError as e:
                        print(e)
                        assert False, "Yaml file has syntax error"

                url = details["servers"][0]["url"]

                print()
                print("   Starting:", name)
                print("   PID:     ", pid)
                print("   Spec:    ", _spec)
                print("   URL:     ", url)
                print("   Cloudmesh UI:     ", url + '/ui') 

                print()

                registry = Registry()
                registry.add_form_file(details,
                                       pid=pid,
                                       spec=_spec,
                                       directory=self.directory,
                                       port=self.port,
                                       host=self.host,
                                       url=url
                                       )

        return pid

    @staticmethod
    def ps(name=None):
        """
        List all of the actively running servers or if name is provided return whether the server is running

        :param name:  Optional server name.  If not provided all servers will be listed.
        :return:  list of pids for all running servers.
        """

        pids = []

        result = Shell.ps()
        #result = result.split('\n')
        for pinfo in result:
            if pinfo["cmdline"] is not None:
                line = ' '.join(pinfo["cmdline"])
                if "openapi server start" in line:
                    print(pinfo)
                    info = line.split("start")[1].split("--")[0].strip()
                    if name is None:
                        name = os.path.basename(
                            line.split("openapi server start")[1]
                        ).split(".")[0]
                        pids.append({"name": name, "pid": pinfo['pid'], "spec": info})
                    elif name is not None and f"{name}.yaml" in info:
                        pids.append({"name":name, "pid": pinfo['pid'], "spec": info})
                    else:
                        pids.append({"name":name, "pid": pinfo["pid"], "spec":info})
                elif "cmsoaserver.py" in line and sys.platform == 'win32':
                    info = line.split("python.exe")[1].strip()
                    if name is None:
                        name = Path(info).stem.split("_")[0].split()
                    pids.append({"name": name, "pid": pinfo['pid'], "spec": info})
        print(pids)
        return pids

    @staticmethod
    def list(name=None):
        """
        Lists the servers that have been registered in Registry

        :param name:  Optional server name.  If not provided all servers in registry will be listed.
        :return:  list of servers in registry
        """
        registry = Registry()
        result = registry.list(name)

        return result

    @staticmethod
    def stop(name=None):
        """
        Stop a running OpenApi server

        :param name:  server name
        :return: 
        """

        Console.ok(f"shutting down server {name}")

        registry = Registry()

        entries = registry.list(name=name)
        pid = ""
        result = None
        if len(entries) > 1:
            Console.error(f"Aborting, returned more than one entry from the Registry with the name {name}")
            raise Exception
        elif len(entries) == 1:
            pid = str(entries[0]['pid'])
        else:
            result = Server.ps(name=name)

        try:
            if len(pid) > 0:  # check if pid found in registry and if found kill
                print("Killing:", pid)
                Shell.kill_pid(pid)
                registry.delete(name=name)
            elif result:  # no server found in registry so kill based on ps output
                for process in result:
                    pid = str(process["pid"])
                    if len(pid) > 0:
                        print("Killing:", pid)
                        Shell.kill_pid(pid)
                        registry.delete(name=name)
                    else:
                        print()
                        Console.error(f"PS output generated invalid PID for server name {name}")
            else:
                print()
                Console.error(f"No Cloudmesh OpenAPI Server found with the name {name}")
        except:
            print("got to exception")
            Console.error(
                f"No Cloudmesh OpenAPI Server found with the name {name}")


    def run_os(self):
        """
        Start an openapi server by creating a physical flask script

        :return:  pid for started server
        """

        Console.ok("starting server")

        # For windows convert backslashes to forward slashes to be python compatible
        if sys.platform == 'win32':
            self.spec = "/".join(self.spec.replace('C:', '').split('\\'))
            self.directory = "/".join(self.directory.replace('C:', '').split('\\'))

        today_dt = date.today().strftime("%m%d%Y")
        VERBOSE("spec path: ", self.spec)

        flask_script = textwrap.dedent(f'''
                import connexion
    
                # Create the application instance
                app = connexion.App(__name__, specification_dir='{self.directory}')
    
                # Read the yaml file to configure the endpoints
                app.add_api('{self.spec}')
    
                if __name__ == '__main__':
                    app.run(host='{self.host}',
                            port={self.port},
                            debug={self.debug},
                            server='{self.server}')
            ''')

        VERBOSE("server script: ", f"{self.directory}/{self.name}_cmsoaserver.py")

        # Write out flask python script to file so that it can be run in background
        try:
            version = open(f"{self.directory}/{self.name}_cmsoaserver.py", 'w').write(
                flask_script)
        except IOError:
            Console.error("Unable to write server file")
        except Exception as e:
            print(e)

        # Run python flask script in background
        pid = ""
        try:
            # TODO: need to write log somewhere else or use a logger to write to common log.  Tried devnull and that does not work.
            logname = f"{self.directory}/{self.name}_server.{today_dt}.log"

            f = open(logname, "w")
            #f = open(os.devnull, "w")
            process = subprocess.Popen([sys.executable,
                                        f"{self.directory}/{self.name}_cmsoaserver.py"],
                                       stdout=f,
                                       stderr=subprocess.STDOUT,
                                       shell=False)

            pid = process.pid

        except Exception as e:
            Console.error("Unable to start server")
            print(e)

        with open(self.spec, "r") as stream:
            try:
                details = yaml.safe_load(stream)
            except yaml.YAMLError as e:
                print(e)
                assert False, "Yaml file has syntax error"

        url = details["servers"][0]["url"]

        print()
        print("   Starting:", self.name)
        print("   PID:     ", pid)
        print("   Spec:    ", self.spec)
        print("   URL:     ", url)
        print()

        registry = Registry()
        registry.add_form_file(details,
                               pid=pid,
                               spec=self.spec,
                               directory=self.directory,
                               port=self.port,
                               host=self.host,
                               url=url
                               )

        return pid




from sklearn.linear_model import RidgeCV
import array
from cloudmesh.openapi.registry.cache import ResultCache


def fit(X: array, y: array, sample_weight: array):

    """
    Fit Ridge regression model with cv.


    :param X: Training data. If using GCV, will be cast to float64
                    if necessary.
    :type X: array
    :param y: Target values. Will be cast to X's dtype if necessary.
    :type y: array
    :param sample_weight: Individual weights for each sample. If given a float, every sample
                    will have the same weight.
    :type sample_weight: array
    :param return: 
    :type return: self
    
    """

    fit = RidgeCV().fit(X, y, sample_weight)
    ResultCache().save("RidgeCV","pickle",fit)


    return RidgeCV


def get_params(deep: bool) -> str:

    """
    Get parameters for this estimator.


    :param deep: If True, will return the parameters for this estimator and
                    contained subobjects that are estimators.
    :type deep: bool
    :param return: Parameter names mapped to their values.
    :type return: str
    
    """

    model = ResultCache().load("RidgeCV")
    str = model.get_params(deep)



    return str


def predict(X: array) -> array:

    """
    Predict using the linear model.


    :param X: Samples.
    :type X: array
    :param return: Returns predicted values.
    :type return: array
    
    """

    model = ResultCache().load("RidgeCV")
    array = model.predict(X)



    return array


def score(X: array, y: array, sample_weight: array) -> float:

    """
    Return the coefficient of determination R^2 of the prediction.


    :param X: Test samples. For some estimators this may be a
                    precomputed kernel matrix or a list of generic objects instead,
                    shape = (n_samples, n_samples_fitted),
                    where n_samples_fitted is the number of
                    samples used in the fitting for the estimator.
    :type X: array
    :param y: True values for X.
    :type y: array
    :param sample_weight: Sample weights.
    :type sample_weight: array
    :param return: R^2 of self.predict(X) wrt. y.
    :type return: float
    
    """

    model = ResultCache().load("RidgeCV")
    float = model.score(X, y, sample_weight)



    return float


def set_params(**params: dict):

    """
    Set the parameters of this estimator.


    :param **params: Estimator parameters.
    :type **params: dict
    :param return: Estimator instance.
    :type return: self
    
    """

    set_params = RidgeCV().set_params(**params)
    ResultCache().save("RidgeCV","pickle",set_params)


    return RidgeCV



from cloudmesh.openapi.scikitlearn import SklearnGenerator

input_sklibrary = 'sklearn.linear_model.LogisticRegression'
model_tag = 'JagsLogis1'
SklearnGenerator.Sklearngenerator(input_sklibrary,model_tag)




import connexion

# Create the application instance
app = connexion.App(__name__, specification_dir='/Users/Jonathan/cm/cloudmesh-openapi/tests/generator')

# Read the yaml file to configure the endpoints
app.add_api('/Users/Jonathan/cm/cloudmesh-openapi/tests/generator/LogisticRegression.yaml')

if __name__ == '__main__':
    app.run(host='127.0.0.1',
            port=8080,
            debug=True,
            server='flask')



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
import time
from pprint import pprint

import pytest
import tests.util as util
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.Shell import Shell
from cloudmesh.common.util import HEADING

directory = "./tests/server-cpu"
filename = "cpu.py"

location = f"{directory}/{filename}"


"cms openapi generate calculator --baseurl="./tests/generator-calculator" --filename=calculator.py --yamldirectory="./tests/generator-calculator" --all_functions"


@pytest.mark.incremental
class TestGenerator:

    def test_generate(self):
        """
        function to validate paths information
        """
        HEADING()

        Benchmark.Start()
        Shell.run(
            "cms openapi generate cpu --yamldirectory={directory} --filename={location}")
        Benchmark.Stop()

    def test_read_spec(self):
        """
        function to check if YAML synatx is correct or not
        """
        global spec

        HEADING()
        Benchmark.Start()
        spec = util.readyaml(location)
        keys = spec.keys()

        pprint(spec)
        pprint(keys)
        assert "openapi" in keys
        assert "info" in keys
        assert "servers" in keys
        assert "paths" in keys
        assert "A simple service" in str(spec)


class rest:

    def test_service(self):
        """
        function to test if the server is started and available to return
        a successful http code
        """
        HEADING()
        Benchmark.Start()

        server_output = Shell.cms("openapi server start {filename}")
        assert server_output.__contains__("starting server")

        time.sleep(2)
        baseurl = "http://127.0.0.1:8080/cloudmesh"

        curl = f"curl {baseurl}/cpu"

        response = Shell.run(curl)
        assert response.__contains__("200")

        Shell.cms("openapi server stop cpu")
        response = Shell.run(curl)
        fail_message = "Failed to connect to 127.0.0.1 port 80: Connection refuse"
        assert response.__contains__(fail_message)

        Benchmark.Stop()

    def test_benchmark(self):
        HEADING()
        Benchmark.print(csv=True, sysinfo=False, tag="generator")



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
import sys
sys.path.append("./tests/lib")
from generator_test import GeneratorBaseTest, ServerBaseTest
from tests.generator import LinearRegression
import pytest

from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.util import HEADING
from cloudmesh.management.configuration.name import Name
service ="openapi"
Benchmark.debug()
X = [1,2,3,4,5,6,7,8]
y = [1,3,5,7]
sample_weight = [1,1,1,1]
X_shape_x = 4
X_shape_y = 2

filename="./tests/generator/LinearRegression.py"
all_functions= True
import_class=False

@pytest.mark.incremental
class TestGeneratorTestClass():


    @pytest.fixture(scope="module")
    def generatorBaseTestFixture(self):
        gen= GeneratorBaseTest(filename,all_functions,import_class)
        return gen

    @pytest.fixture(scope="module")
    def serverBaseTestFixture(self):
        server = ServerBaseTest()
        return server


    def test_copy_file(self,generatorBaseTestFixture):
        generatorBaseTestFixture.copy_py_file()

    def test_generate(self,generatorBaseTestFixture):
        """
        function to validate paths information
        """
        generatorBaseTestFixture.generate()

    def test_read_spec(self,generatorBaseTestFixture):
        generatorBaseTestFixture.read_spec()

    def test_validate_function(self, generatorBaseTestFixture):
        generatorBaseTestFixture.validate_function()

    def test_fit(self):
        """
        function to test if the server is started and available to return
        a successful http code
        """

        HEADING()
        Benchmark.Start()
        LinearRegression.fit(X, y, sample_weight, X_shape_x, X_shape_y)
        Benchmark.Stop()
        assert True

    def test_score(self):
        """
        function to test the score
        """
        HEADING()
        Benchmark.Start()
        score = LinearRegression.score(X,y,sample_weight,X_shape_x,X_shape_y)
        Benchmark.Stop()
        assert score > 0

    def test_predict(self):
        """
        function to test the predict
        """
        HEADING()
        Benchmark.Start()
        LinearRegression.predict(X,X_shape_x,X_shape_y)
        Benchmark.Stop()
        assert True

    def test_start_service(self,serverBaseTestFixture):
        serverBaseTestFixture.start_service()

    def test_stop_server(self,serverBaseTestFixture):
        serverBaseTestFixture.stop_server()

    def test_delete_build_file(self, generatorBaseTestFixture):
        generatorBaseTestFixture.delete_file()

    def test_benchmark(self):
        Benchmark.print(sysinfo=True, csv=True, tag=service)




import os
import os
import sys
import textwrap
import types
from dataclasses import is_dataclass
from importlib import import_module
from pathlib import Path
from shutil import copyfile

from cloudmesh.common.Printer import Printer
from cloudmesh.common.console import Console
from cloudmesh.common.debug import VERBOSE
from cloudmesh.common.util import path_expand
from cloudmesh.openapi.authentication.basic import BasicAuth
from cloudmesh.openapi.function import generator
from cloudmesh.openapi.function.server import Server
from cloudmesh.openapi.function.executor import Parameter
from cloudmesh.openapi.registry.Registry import Registry
from cloudmesh.openapi.scikitlearn.SklearnGenerator import Sklearngenerator
from cloudmesh.openapi.scikitlearn.SklearnGeneratorFile import SklearngeneratorFile
from cloudmesh.shell.command import PluginCommand
from cloudmesh.shell.command import command, map_parameters


# start-stop: osx Andrew
# start_stop: windows Jonathan
# start-stop: linux Prateek


class OpenapiCommand(PluginCommand):

    # noinspection PyUnusedLocal,PyPep8Naming
    @command
    def do_openapi(self, args, arguments):
        """
        ::

          Usage:
              openapi generate [FUNCTION] --filename=FILENAME
                                         [--serverurl=SERVERURL]
                                         [--yamlfile=YAML]
                                         [--import_class]
                                         [--all_functions]
                                         [--enable_upload]
                                         [--verbose]
                                         [--basic_auth=CREDENTIALS]
              openapi server start YAML [NAME]
                              [--directory=DIRECTORY]
                              [--port=PORT]
                              [--server=SERVER]
                              [--host=HOST]
                              [--verbose]
                              [--debug]
                              [--fg]
                              [--os]
              openapi server stop NAME
              openapi server list [NAME] [--output=OUTPUT]
              openapi server ps [NAME] [--output=OUTPUT]
              openapi register add NAME ENDPOINT
              openapi register filename NAME
              openapi register delete NAME
              openapi register list [NAME] [--output=OUTPUT]
              openapi TODO merge [SERVICES...] [--dir=DIR] [--verbose]
              openapi TODO doc FILE --format=(txt|md)[--indent=INDENT]
              openapi TODO doc [SERVICES...] [--dir=DIR]
              openapi sklearn FUNCTION MODELTAG
              openapi sklearnreadfile FUNCTION MODELTAG
              openapi sklearn upload --filename=FILENAME

          Arguments:
              FUNCTION  The name for the function or class
              MODELTAG  The arbirtary name choosen by the user to store the Sklearn trained model as Pickle object
              FILENAME  Path to python file containing the function or class
              SERVERURL OpenAPI server URL Default: https://localhost:8080/cloudmesh
              YAML      Path to yaml file that will contain OpenAPI spec. Default: FILENAME with .py replaced by .yaml
              DIR       The directory of the specifications
              FILE      The specification

          Options:
              --import_class         FUNCTION is a required class name instead of an optional function name
              --all_functions        Generate OpenAPI spec for all functions in FILENAME
              --debug                Use the server in debug mode
              --verbose              Specifies to run in debug mode
                                     [default: False]
              --port=PORT            The port for the server [default: 8080]
              --directory=DIRECTORY  The directory in which the server is run
              --server=SERVER        The server [default: flask]
              --output=OUTPUT        The outputformat, table, csv, yaml, json
                                     [default: table]
              --srcdir=SRCDIR        The directory of the specifications
              --destdir=DESTDIR      The directory where the generated code
                                     is placed

          Description:
            This command does some useful things.

            openapi TODO doc FILE --format=(txt|md|rst) [--indent=INDENT]
                Sometimes it is useful to generate teh openaopi documentation
                in another format. We provide fucntionality to generate the
                documentation from the yaml file in a different formt.

            openapi TODO doc --format=(txt|md|rst) [SERVICES...]
                Creates a short documentation from services registered in the
                registry.

            openapi TODO merge [SERVICES...] [--dir=DIR] [--verbose]
                Merges tow service specifications into a single servoce
                TODO: do we have a prototype of this?


            openapi sklearn sklearn.linear_model.LogisticRegression
                Generates the .py file for the Model given for the generator

            openapi sklearnreadfile sklearn.linear_model.LogisticRegression
            Generates the .py file for the Model given for the generator which supports reading files

            openapi generate [FUNCTION] --filename=FILENAME
                                         [--serverurl=SERVERURL]
                                         [--yamlfile=YAML]
                                         [--import_class]
                                         [--all_functions]
                                         [--enable_upload]
                                         [--verbose]
                                         [--basic_auth=CREDENTIALS]
                Generates an OpenAPI specification for FUNCTION in FILENAME and
                writes the result to YAML. Use --import_class to import a class
                with its associated class methods, or use --all_functions to 
                import all functions in FILENAME. These options ignore functions
                whose names start with '_'. Use --enable_upload to add file
                upload functionality to a copy of your python file and the
                resulting yaml file.

                For optional basic authorization, we support (temporarily) a single user
                credential. CREDENTIALS should be formatted as follows:

                user:password

                Example: --basic_auth=admin:secret

            openapi server start YAML [NAME]
                              [--directory=DIRECTORY]
                              [--port=PORT]
                              [--server=SERVER]
                              [--host=HOST]
                              [--verbose]
                              [--debug]
                              [--fg]
                              [--os]
                starts an openapi web service using YAML as a specification
                TODO: directory is hard coded as None, and in server.py it
                  defaults to the directory where the yaml file lives. Can
                  we just remove this argument?

            openapi server stop NAME
                stops the openapi service with the given name
                TODO: where does this command has to be started from

            openapi server list [NAME] [--output=OUTPUT]
                Provides a list of all OpenAPI services in the registry

            openapi server ps [NAME] [--output=OUTPUT]
                list the running openapi service

            openapi register add NAME ENDPOINT
                Openapi comes with a service registry in which we can register
                openapi services.

            openapi register filename NAME
                In case you have a yaml file the openapi service can also be
                registerd from a yaml file

            openapi register delete NAME
                Deletes the names service from the registry

            openapi register list [NAME] [--output=OUTPUT]
                Provides a list of all registerd OpenAPI services


        """
        #print(arguments)
        map_parameters(arguments,
                       'fg',
                       'os',
                       'output',
                       'verbose',
                       'port',
                       'directory',
                       'yamlfile',
                       'serverurl',
                       'name',
                       'import_class',
                       'all_functions',
                       'enable_upload',
                       'host',
                       'basic_auth')
        arguments.debug = arguments.verbose

        #VERBOSE(arguments)

        if arguments.generate:
            if arguments.import_class and arguments.all_functions:
                Console.error('Cannot generate openapi with both --import_class and --all_functions')
            if arguments.import_class and not arguments.FUNCTION:
                Console.error('FUNCTION parameter (class name) is required when using --import_class')
            try:
                p = Parameter(arguments)
                p.Print()
                filename = p.filename # ./dir/myfile.py
                yamlfile = p.yamlfile # ./dir/myfile.yaml
                directory = p.yamldirectory # ./dir
                function = p.function # myfunction
                serverurl = p.serverurl # http://localhost:8080/cloudmesh/
                module_name = p.module_name # myfile
                basic_auth = p.basic_auth # user:password

                # If statement here for mode with basic_auth

                enable_upload = arguments.enable_upload
                # append the upload function to the end of a copy of the file if not already done
                if enable_upload:
                    uploadPython = textwrap.dedent("""
                        from cloudmesh.openapi.registry.fileoperation import FileOperation
                        
                        def upload() -> str:
                            filename=FileOperation().file_upload()
                            return filename
                        
                        #### upload functionality added
                        """)
                    upload_added = False
                    for line in open(filename):
                        if '#### upload functionality added' in line:
                            upload_added = True
                    if not upload_added:
                        filename_upload = filename.replace('.py', '_upload-enabled.py')
                        copyfile(filename, filename_upload)
                        Console.info(f'copied {filename} to {filename_upload}')
                        filename = filename_upload
                        module_name = module_name + '_upload-enabled'
                        with open(filename, 'a') as f:
                            f.write('\n')
                            f.write(uploadPython)
                        Console.info(f'added upload functionality to {filename}')

                if basic_auth:
                    user, password = basic_auth.split(':')
                    BasicAuth.reset_users()
                    BasicAuth.add_user(user, password)
                    module_name, filename = BasicAuth.write_basic_auth(
                        filename=filename,
                         module_name=module_name)

                # Parameter() takes care of putting the filename in the path
                imported_module = import_module(module_name)
                dataclass_list = []
                for attr_name in dir(imported_module):
                    attr = getattr(imported_module, attr_name)
                    if is_dataclass(attr):
                        dataclass_list.append(attr)
                # not currently supporting multiple functions or all functions
                # could do comma-separated function/class names

                if enable_upload:
                    upload_obj = getattr(imported_module, 'upload')
                    setattr(sys.modules[module_name], 'upload', upload_obj)

                if arguments.import_class:
                    class_obj = getattr(imported_module, function)
                    # do we maybe need to do this here?
                    # setattr(sys.modules[module_name], function, class_obj)
                    class_description = class_obj.__doc__.strip().split("\n")[0]
                    func_objects = {}
                    for attr_name in dir(class_obj):
                        attr = getattr(class_obj, attr_name)                        
                        if isinstance(attr, types.MethodType) and attr_name[0] != '_':
                            # are we sure this is right?
                            # would probably create a valid openapi yaml, but not technically accurate
                            # module.function may work but it should be module.Class.function
                            setattr(sys.modules[module_name], attr_name, attr)
                            func_objects[attr_name] = attr
                        elif is_dataclass(attr):
                            dataclass_list.append(attr)
                    openAPI = generator.Generator()
                    Console.info(
                        'Generating openapi for class: ' + class_obj.__name__)
                    openAPI.generate_openapi_class(class_name=class_obj.__name__,
                                                   class_description=class_description,
                                                   filename=filename,
                                                   func_objects=func_objects,
                                                   serverurl=serverurl,
                                                   outdir=directory,
                                                   yamlfile=yamlfile,
                                                   dataclass_list=dataclass_list,
                                                   all_function=False,
                                                   enable_upload=enable_upload,
                                                   basic_auth_enabled=basic_auth,
                                                   write=True)
                elif arguments.all_functions:
                    func_objects = {}
                    for attr_name in dir(imported_module):
                        if type(getattr(imported_module, attr_name)).__name__ == 'function' and attr_name[0] != '_':
                            func_obj = getattr(imported_module, attr_name)
                            setattr(sys.modules[module_name], attr_name, func_obj)
                            func_objects[attr_name] = func_obj
                    openAPI = generator.Generator()
                    Console.info(
                        'Generating openapi for all functions in file: ' + filename)
                    openAPI.generate_openapi_class(class_name=module_name,
                                                   class_description="No description provided",
                                                   filename=filename,
                                                   func_objects=func_objects,
                                                   serverurl=serverurl,
                                                   outdir=directory,
                                                   yamlfile=yamlfile,
                                                   dataclass_list=dataclass_list,
                                                   all_function=True,
                                                   enable_upload=enable_upload,
                                                   basic_auth_enabled=basic_auth,
                                                   write=True)

                else:
                    func_obj = getattr(imported_module, function)
                    setattr(sys.modules[module_name], function, func_obj)
                    openAPI = generator.Generator()
                    Console.info(
                        'Generating openapi for function: ' + func_obj.__name__)
                    openAPI.generate_openapi(f=func_obj,
                                             filename=filename,
                                             serverurl=serverurl,
                                             outdir=directory,
                                             yamlfile=yamlfile,
                                             dataclass_list=dataclass_list,
                                             enable_upload=enable_upload,
                                             basic_auth_enabled=basic_auth,
                                             write=True)

            except Exception as e:
                Console.error("Failed to generate openapi yaml")
                print(e)

        elif arguments.server and arguments.start and arguments.os:

            try:
                s = Server(
                    name=arguments.NAME,
                    spec=path_expand(arguments.YAML),
                    directory=path_expand(
                        arguments.directory) or arguments.directory,
                    port=arguments.port,
                    server=arguments.wsgi,
                    debug=arguments.debug
                )

                pid = s.run_os()

                VERBOSE(arguments, label="Server parameters")

                print(f"Run PID: {pid}")

            except FileNotFoundError:

                Console.error("specification file not found")

            except Exception as e:

                print(e)

        elif arguments.server and arguments.list:

            try:
                result = Server.list(name=arguments.NAME)

                # BUG: order= nt yet defined

                print(Printer.list(result))

            except ConnectionError:
                Console.error("Server not running")

        elif arguments.server and arguments.ps:

            try:
                print()
                Console.info("Running Cloudmesh OpenAPI Servers")
                print()
                result = Server.ps(name=arguments.NAME)
                print(Printer.list(result, order=["name", "pid", "spec"]))

                print()
            except ConnectionError:
                Console.error("Server not running")

        elif arguments.register and arguments.add:

            registry = Registry()
            result = registry.add(name=arguments.NAME, url=arguments.BASEURL,
                                  pid=arguments.PID)

            registry.Print(data=result, output=arguments.output)

        elif arguments.register and arguments.delete:

            registry = Registry()
            result = registry.delete(name=arguments.NAME)
            if result == 0:
                Console.error("Entry could not be found")
            else:
                Console.ok("Ok. Entry deleted")

        elif arguments.register and arguments.list:

            registry = Registry()
            result = registry.list(name=arguments.NAME)

            registry.Print(data=result, output=arguments.output)

        elif arguments.register and arguments['filename']:

            registry = Registry()
            result = [registry.add_form_file(arguments['filename'])]

            registry.Print(data=result, output=arguments.output)

        elif arguments.server and arguments.start:

            # VERBOSE(arguments)

            try:
                s = Server(
                    name=arguments.NAME,
                    spec=path_expand(arguments.YAML),
                    directory=None,
                    #directory=path_expand(
                    #    arguments.directory) or arguments.directory,
                    port=arguments.port,
                    host=arguments.host,
                    server=arguments.wsgi,
                    debug=arguments.debug)

                pid = s.start(name=arguments.NAME,
                              spec=path_expand(arguments.YAML),
                              foreground=arguments.fg)

                print(f"Run PID: {pid}")

            except FileNotFoundError:

                Console.error("specification file not found")

            except Exception as e:
                print(e)

        elif arguments.server and arguments.stop:

            try:
                print()
                Console.info("Stopping Cloudmesh OpenAPI Server")
                print()

                Server.stop(name=arguments.NAME)

                print()
            except ConnectionError:
                Console.error("Server not running")

        elif arguments.sklearn and not arguments.upload:

            try:
                Sklearngenerator(input_sklibrary=arguments.FUNCTION,
                                 model_tag=arguments.MODELTAG)
            except Exception as e:
                print(e)

        elif arguments.sklearnreadfile and not arguments.upload:

            try:
                SklearngeneratorFile(input_sklibrary=arguments.FUNCTION,
                                          model_tag=arguments.MODELTAG)
            except Exception as e:
                print(e)
        
        #TODO: implement this?
        elif arguments.sklearn and arguments.upload:

            try:
                openAPI = generator.Generator()
                openAPI.fileput()

            except Exception as e:
                print(e)

        '''

        arguments.wsgi = arguments["--server"]

        m = Manager(debug=arguments.debug)

        arguments.dir = path_expand(arguments["--dir"] or ".")

        # pprint(arguments)

        if arguments.list:
            files = m.get(arguments.dir)
            print("List of specifications")
            print(79 * "=")
            print('\n'.join(files))

        elif arguments.merge:
            d = m.merge(arguments.dir, arguments.SERVICES)
            print(yaml.dump(d, default_flow_style=False))

        elif arguments.description:
            d = m.description(arguments.dir, arguments.SERVICES)

        elif arguments.md:

            converter = OpenAPIMarkdown()

            if arguments["--indent"] is None:
                indent = 1
            else:
                indent = int(arguments["--indent"])
            filename = arguments["FILE"]

            converter.title(filename, indent=indent)
            converter.convert_definitions(filename, indent=indent + 1)
            converter.convert_paths(filename, indent=indent + 1)
        elif arguments.codegen:
            m.codegen(arguments.SERVICES, arguments.dir)


        return ""
        '''




from sklearn.linear_model import LogisticRegression
import array
from cloudmesh.openapi.registry.cache import ResultCache


def decision_function(X: array) -> array:

    """
    Predict confidence scores for samples.


    :param X: Samples.
    :type X: array
    :param return: Confidence scores per (sample, class) combination. In the binary
                    case, confidence score for self.classes_[1] where >0 means this
                    class would be predicted.
    :type return: array
    
    """

    model = ResultCache().load("LogisticRegression")
    array = model.decision_function(X)



    return array


def densify():

    """
    Convert coefficient matrix to dense array format.


    :param return: Fitted estimator.
    :type return: self
    
    """

    densify = densify()



    return densify


def fit(X: str, y: array, sample_weight: array):

    """
    Fit the model according to the given training data.


    :param X: Training vector, where n_samples is the number of samples and
                    n_features is the number of features.
    :type X: str
    :param y: Target vector relative to X.
    :type y: array
    :param sample_weight: Array of weights that are assigned to individual samples.
                    If not provided, then each sample is given unit weight.
                    
                    .. versionadded:: 0.17
                       *sample_weight* support to LogisticRegression.
    :type sample_weight: array
    :param return: Fitted estimator.
    :type return: self
    
    """

    fit = LogisticRegression().fit(X, y, sample_weight)
    ResultCache().save("LogisticRegression","pickle",fit)


    return LogisticRegression


def get_params(deep: bool) -> str:

    """
    Get parameters for this estimator.


    :param deep: If True, will return the parameters for this estimator and
                    contained subobjects that are estimators.
    :type deep: bool
    :param return: Parameter names mapped to their values.
    :type return: str
    
    """

    model = ResultCache().load("LogisticRegression")
    str = model.get_params(deep)



    return str


def predict(X: array) -> array:

    """
    Predict class labels for samples in X.


    :param X: Samples.
    :type X: array
    :param return: Predicted class label per sample.
    :type return: array
    
    """

    model = ResultCache().load("LogisticRegression")
    array = model.predict(X)



    return array


def predict_log_proba(X: array) -> array:

    """
    Predict logarithm of probability estimates.


    :param X: Vector to be scored, where `n_samples` is the number of samples and
                    `n_features` is the number of features.
    :type X: array
    :param return: Returns the log-probability of the sample for each class in the
                    model, where classes are ordered as they are in ``self.classes_``.
    :type return: array
    
    """

    model = ResultCache().load("LogisticRegression")
    array = model.predict_log_proba(X)



    return array


def predict_proba(X: array) -> array:

    """
    Probability estimates.


    :param X: Vector to be scored, where `n_samples` is the number of samples and
                    `n_features` is the number of features.
    :type X: array
    :param return: Returns the probability of the sample for each class in the model,
                    where classes are ordered as they are in ``self.classes_``.
    :type return: array
    
    """

    model = ResultCache().load("LogisticRegression")
    array = model.predict_proba(X)



    return array


def score(X: array, y: array, sample_weight: array) -> float:

    """
    Return the mean accuracy on the given test data and labels.


    :param X: Test samples.
    :type X: array
    :param y: True labels for X.
    :type y: array
    :param sample_weight: Sample weights.
    :type sample_weight: array
    :param return: Mean accuracy of self.predict(X) wrt. y.
    :type return: float
    
    """

    model = ResultCache().load("LogisticRegression")
    float = model.score(X, y, sample_weight)



    return float


def set_params(**params: dict):

    """
    Set the parameters of this estimator.


    :param **params: Estimator parameters.
    :type **params: dict
    :param return: Estimator instance.
    :type return: self
    
    """

    set_params = LogisticRegression().set_params(**params)
    ResultCache().save("LogisticRegression","pickle",set_params)


    return LogisticRegression


def sparsify():

    """
    Convert coefficient matrix to sparse format.


    :param return: Fitted estimator.
    :type return: self
    
    """

    sparsify = sparsify()



    return sparsify



import oyaml as yaml
from cloudmesh.common.util import path_expand



def readyaml(name):
    with open(path_expand(name), "r") as stream:
        try:
            spec = yaml.safe_load(stream)
        except yaml.YAMLError as e:
            print(e)
            assert False, "Yaml file has syntax error"
    return spec



import pathlib
import textwrap
import re
from dataclasses import is_dataclass

import requests
from cloudmesh.common.console import Console
from cloudmesh.common.debug import VERBOSE
from docstring_parser import parse


# TODO: Use latest version of openapi
# TODO: Use a dynamic way to derive the version of module/class being generated

class Generator:

    basicAuthTemplate = textwrap.dedent("""
          securitySchemes:
            basicAuth: # <-- arbitrary name for the security scheme
              type: http
              scheme: basic
              x-basicInfoFunc: {filename}.basic_auth
        
        security:
          - basicAuth: [] # <-- use the same name here
        """)

    openAPITemplate = textwrap.dedent("""
        openapi: 3.0.0
        info:
          title: {title}
          description: "{description}"
          version: "{version}"
        servers:
          - url: {serverurl}
            description: "{description}"
        paths:
          /{name}:
             get:
              summary: "{description}"
              description: Optional extended description in CommonMark or HTML.
              operationId: {filename}.{name}
              parameters:
                {parameters}
              responses:
                {responses}
        {upload}
        {components}
        """)

    openAPITemplate2 = textwrap.dedent("""
        openapi: 3.0.0
        info:
          title: {title}
          description: "{description}"
          version: "{version}"
        servers:
          - url: {serverurl}
            description: "{description}"
        paths:
          {paths}
        {upload}
        {components}
        """)

    uploadTemplate = textwrap.dedent("""
        /upload:
          post:
            summary: upload a file
            operationId: {filename}.upload
            requestBody:
              content:
                multipart/form-data:
                  schema:
                    type: object
                    properties:
                      upload:
                        type: string
                        format: binary
            responses:
              '200':
                description: "OK"
                content:
                  text/plain:
                    schema:
                      type: string
                      """)

    def parse_type(self, _type):
        """
        Function to lookup and output supported OpenApi data type using python data type as input

        :param _type:  the input data type that will be converted to an openapi compliant data type
        :return:  Returns openapi compliant data type
        """

        parser = {
            'int': 'type: integer',
            'bool': 'type: boolean',
            'float': 'type: number',
            'str': 'type: string',
            'list': 'type: array',
            'array': 'type: array',
            'dict': 'type: object'
        }

        if is_dataclass(_type):
            return f'$ref: "#/components/schemas/{_type}'
        # exits with KeyError if unsupported type is given

        try:
            t = parser[_type]
        except KeyError:
            print(f'unsupported data type supplied for {_type}')
            raise Exception
        return t

    def generate_parameter(self, name, _type, description):
        """
        Function to generate a single OpenApi YAML formatted parameter section

        :param name:   input python function parameter name
        :param _type:  input python function parameter type
        :param description:  input python function parameter description
        :return:  parameter spec section
        """

        # Note: did not use f string approach to populate parameter values in strings due to indentation issues.
        #   It seems that f string approach interprets newlines in the parameter values after embedding in string and
        #   we need it to happen before.

        if type(_type) == str:
            _type = self.parse_type(_type)
        else:
            _type = self.parse_type(_type.__name__)

        if _type.find('array') != -1:
            spec = textwrap.dedent("""
                - in: query
                  name: {name}
                  description: "{description}"
                  schema:
                    {_type}
                    items: 
                      type: number""").format(name=name.strip(),
                                       description=description.strip(),
                                       _type=_type.strip())
        elif _type.find('object') != -1:
            spec = textwrap.dedent("""
                - in: query
                  name: {name}
                  description: "{description}"
                  schema:
                    {_type}
                    additionalProperties: true""").format(name=name.strip(),
                                       description=description.strip(),
                                       _type=_type.strip())
        else:
            spec = textwrap.dedent("""
                - in: query
                  name: {name}
                  description: "{description}"
                  schema:
                    {_type}""").format(name=name.strip(),
                                       description=description.strip(),
                                       _type=_type.strip())

        return spec

    def generate_response(self, code, _type, description):
        """
        Function to generate a single OpenApi YAML formatted response section

        :param code:  openapi response code
        :param _type:  openapi response type
        :param description:  openapi response decription
        :return:  openapi response section spec
        """

        if type(_type) == str:
            # Only retrieve openapi data type for responses that return a value
            if _type == "No Response":
                VERBOSE("Processing operation with no response")
            else:
                _type = self.parse_type(_type)
        else:
            _type = self.parse_type(_type.__name__)

        if _type == "No Response":
            spec = textwrap.dedent("""
              '{code}':
                description: "{description}"
            """).format(code=code.strip(),
                        description=description.strip()
                        )
        elif not _type.startswith('object'):
            # int, bool, float, str, list
            if (_type.find('array') != -1):
                spec = textwrap.dedent("""
                  '{code}':
                    description: "{description}"
                    content:
                      text/plain:
                        schema:
                          {_type}
                          items: {{}}""").format(code=code.strip(),
                                             description=description.strip(),
                                             _type=_type.strip())
            else:
                spec = textwrap.dedent("""
                  '{code}':
                    description: "{description}"
                    content:
                      text/plain:
                        schema:
                          {_type}""").format(code=code.strip(),
                                             description=description.strip(),
                                             _type=_type.strip())
        else:
            # dict (generic json) or dataclass ($ref)
            if (_type.find('object') != -1):
                spec = textwrap.dedent("""
                  '{code}':
                    description: "{description}"
                    content:
                      application/json:
                        schema:
                          {_type}
                          additionalProperties: True""").format(code=code.strip(),
                                             description=description.strip(),
                                             _type=_type.strip())
            else:
                spec = textwrap.dedent("""
                  '{code}':
                    description: "{description}"
                    content:
                      application/json:
                        schema:
                          {_type}""").format(code=code.strip(),
                                             description=description.strip(),
                                             _type=_type.strip())
        return spec

    def generate_properties(self, attr, _type):
        """
        Function to generate a single OpenApi YAML formatted schema properties section

        :param attr:  openapi schema property attribute name
        :param _type:  openapi schema property attribute type
        :return:  openapi schema property spec
        """

        if type(_type) == str:
            _type = self.parse_type(_type)
        else:
            _type = self.parse_type(_type.__name__)

        spec = textwrap.dedent(f"""
          {attr}:
            {_type}""")
        return spec

    def generate_schema(self, _class):
        """
        Function to generate a single OpenApi YAML formatted schema section using python dataclass as input

        :param _class:  python type class object
        :return:  openapi schema section spec
        """
        class_name = _class.__name__
        if not is_dataclass(_class):
            raise TypeError(
                f'{class_name} is not a dataclass. '
                'Use the @dataclass decorator to define the class properly')
        properties = str()
        for attr, _type in _class.__annotations__.items():
            properties = properties + self.generate_properties(attr, _type)
        spec = textwrap.dedent(f"""
          {class_name}:
            type: object
            properties:
              {properties}""")
        return spec

    def populate_parameters(self, func_obj):
        """
        Function that converts all the input parameters of a python function into a single OpenApi YAML formatted
        parameters section.

        :param func_obj:  python function ojbect
        :return:  openapi parameters section spec
        """
        spec = str()
        description = None
        for parameter, _type in func_obj.__annotations__.items():
            if parameter == 'return':
                continue  # dicts are unordered, so use continue
                # intead of break to be safe
            else:
                docstring = parse(func_obj.__doc__)
                for param in docstring.params:
                    if param.arg_name == parameter:
                        description = textwrap.indent(param.description, ' ' * 15)
                spec = spec + self.generate_parameter(
                    parameter,
                    _type,
                    description if description else "no description provided in docstring")
                VERBOSE(spec)

        return spec

    def generate_path(self,
                      class_name=None,
                      description=None,
                      long_description=None,
                      funcname=None,
                      parameters=None,
                      responses=None,
                      filename=None,
                      all_function=None):
        """
        Function that generates a single OpenApi YAML formatted operation ID section

        :param class_name:  python class name or module name
        :param description:  python class description
        :param long_description:  pytyhon class long description
        :param funcname:  python function name
        :param parameters:  openapi formatted parameters section spec
        :param responses:  openapi formatted responses section spec
        :param filename:  python module file name
        :param all_function:  all_function flag that means we process all functions in input file
        :return:  openapi operation id section spec
        """

        l_description = long_description \
            if long_description != None \
            else 'None (Optional extended description in CommonMark or HTML)'

        if all_function:
            operationId = f"{filename}.{funcname}"
        else:
            operationId = f"{filename}.{class_name}.{funcname}"

        spec = textwrap.dedent("""
            /{class_name}/{funcname}:
               get:
                summary: "{description}"
                description: "{l_description}"
                operationId: {operationId}
                parameters:
                  {parameters}
                responses:
                  {responses}
        """).format(
            description=description,
            l_description=l_description.strip(),
            class_name=class_name,
            funcname=funcname,
            parameters=parameters.strip(),
            responses=responses.strip(),
            operationId=operationId
        )
        # remove 'parameters:' section if empty
        if parameters == '':
            spec = re.sub('\s*parameters:', '', spec)
        return spec

    def generate_openapi_class(self,
                               class_name=None,
                               class_description=None,
                               filename=None,
                               func_objects=None,
                               serverurl=None,
                               outdir=None,
                               yamlfile=None,
                               dataclass_list=None,
                               all_function=False,
                               enable_upload=False,
                               basic_auth_enabled=False,
                               write=True):
        """
        This is a main entry point into the module.  This function will generate the full OpenApi YAML formatted
        specification for a python class or module with multiple functions.

        :param class_name:  python class name
        :param class_description:  python class description
        :param filename:  python module file name
        :param func_objects:  all function objects in input python file
        :param serverurl:  url for flask service
        :param outdir:  output directory where openapi spec yaml will be written
        :param yamlfile:  file name for openapi spec yaml
        :param dataclass_list:  list containing all data class objects
        :param all_function:  flag to process all functions in input python file.  False by default.
        :param enable_upload:  flag to support file upload function. False by default. 
        :param write:  flag to write out openapi spec yaml to a file. True by default.
        :return:  no return but if write flag is set it will write out yaml to file
        """

        # Initializing and setting global variables
        paths = ""
        description = class_description \
            if class_description \
            else "No description found"
        version = "1.0"  # TODO:  hard coded for now
        filename = pathlib.Path(filename).stem

        # Loop through all functions
        for k, v in func_objects.items():  # k = function_name, v = function object
            VERBOSE(v)
            func_name = v.__name__
            # otherwise it will process the upload function, which doesn't have a defined response
            if func_name == 'upload' and enable_upload == True:
                continue
            Console.info('*'*40)
            Console.info(f"Currently processing function: {func_name}")

            docstring = parse(v.__doc__)

            func_description = docstring.short_description

            func_ldescription = docstring.long_description
            if func_ldescription:
                func_ldescription = textwrap.indent(func_ldescription.strip(), ' ' * 17)

            VERBOSE(func_description)
            VERBOSE(func_ldescription)

            # Define parameters section(s) for openapi yaml
            parameters = self.populate_parameters(v)
            if parameters != "":
                # Console.info(f"Processing parameters for function {func_name}")
                parameters = textwrap.indent(parameters, ' ' * 6)
                VERBOSE(parameters, label="openapi function parameters")
            else:
                Console.info(f"Function {func_name} has no parameters defined.")

            # Define responses section(s) for openapi yaml
            if 'return' in v.__annotations__:
                # Console.info(f"Processing response for function {func_name}")
                responses = self.generate_response('200',
                                                   v.__annotations__['return'],
                                                   'OK')
            else:
                # Console.info(f"Processing NO response for function {func_name}")
                responses = self.generate_response('204',
                                                   "No Response",
                                                   'This operation returns no response.')

            responses = textwrap.indent(responses, ' ' * 6)
            VERBOSE(responses, label="openapi function responses")

            # Define paths section(s) for openapi yaml
            paths = paths + self.generate_path(class_name,
                                               func_description,
                                               func_ldescription,
                                               func_name,
                                               parameters,
                                               responses,
                                               filename,
                                               all_function)

            VERBOSE(paths, label="openapi function path")

        # Indent full paths section for openapi yaml
        paths = textwrap.indent(paths, ' ' * 2)
        VERBOSE(paths, label="openapi function paths")

        # Define components section(s) for openapi yaml
        components = ""
        schemas = ""
        if len(dataclass_list) > 0:
            components = textwrap.dedent("""
                      components:
                        schemas:
                          """)
            for dc in dataclass_list:
                schemas = schemas + textwrap.indent(self.generate_schema(dc), ' ' * 6)
        VERBOSE(components, label="openapi function components")

        upload = ''
        if enable_upload:
            upload = self.uploadTemplate.format(filename=filename)
            upload = textwrap.indent(upload, ' ' * 2)

        if basic_auth_enabled and len(dataclass_list) == 0:
          components = "components:" + self.basicAuthTemplate.format(filename=filename)
        elif basic_auth_enabled:
          components = components + self.basicAuthTemplate.format(filename=filename)

        # Update openapi template to create final version of openapi yaml
        spec = self.openAPITemplate2.format(
            title=class_name,
            description=description,
            version=version,
            paths=paths.strip(),
            serverurl=serverurl,
            filename=filename,
            upload=upload,
            components=components.strip()
        )

        # Write openapi yaml to file
        if write:
            try:
                if yamlfile != "" and yamlfile is not None:
                    version = open(yamlfile, 'w').write(spec.strip())
                else:  # should really never get here
                    version = open(f"{outdir}/{class_name}.yaml", 'w').write(spec.strip())
            except IOError:
                Console.error("Unable to write yaml file")
            except Exception as e:
                print(e)

        return

    def generate_openapi(self,
                         f=None,
                         filename=None,
                         serverurl=None,
                         outdir=None,
                         yamlfile=None,
                         dataclass_list=None,
                         enable_upload=False,
                         basic_auth_enabled=False,
                         write=True):
        """
        This is a main entry point into the module.  This function will generate the full OpenApi YAML formatted
        specification for a module with one single function.

        :param f:  python function object
        :param filename:  python module file name
        :param serverurl:  server url for flask service
        :param outdir:  output directory where openapi spec yaml will be written to
        :param yamlfile:  openapi spec yaml file name
        :param dataclass_list:  list containing all data class objects
        :param write:  flag to write out openapi spec yaml to a file. True by default.
        :return:  no return but if write flag is set it will write out yaml to file.
        """  

        description = f.__doc__.strip().split("\n")[0]
        version = "1.0"  # TODO:  hard coded for now
        title = f.__name__
        parameters = self.populate_parameters(f)
        if parameters != "":
            parameters = textwrap.indent(parameters, ' ' * 8)
            VERBOSE(parameters, label="openapi function parameters")
        else:
            Console.info(f"Function {title} has no parameters defined")

        if 'return' in f.__annotations__:
            # Console.info(f"Processing response for function {title}")
            responses = self.generate_response('200',
                                               f.__annotations__['return'],
                                               'OK')
        else:
            # Console.info(f"Processing NO response for function {title}")
            responses = self.generate_response('204',
                                               "No Response",
                                               'This operation returns no response.')
        responses = textwrap.indent(responses, ' ' * 8)
        VERBOSE(responses, label="openapi function responses")

        components = ''
        schemas = ''
        if len(dataclass_list) > 0:
            components = textwrap.dedent("""
              components:
                schemas:
                  """)
            for dc in dataclass_list:
                schemas = schemas + textwrap.indent(self.generate_schema(dc), ' ' * 6)

        VERBOSE(components, label="openapi function components")

        filename = pathlib.Path(filename).stem
        upload = ''
        if enable_upload:
            upload = self.uploadTemplate.format(filename=filename)
            upload = textwrap.indent(upload, ' ' * 2)

        if basic_auth_enabled and len(dataclass_list) == 0:
          components = "components:" + self.basicAuthTemplate.format(filename=filename)
        elif basic_auth_enabled:
          components = components + self.basicAuthTemplate.format(filename=filename)

        spec = self.openAPITemplate.format(
            title=title,
            name=f.__name__,
            description=description,
            version=version,
            parameters=parameters.strip(),
            responses=responses.strip(),
            serverurl=serverurl,
            filename=filename,
            upload=upload,
            components=components
        )

        # remove 'parameters:' section if empty
        if parameters == '':
            spec = re.sub('\s*parameters:', '', spec)
        if write:
            try:
                if yamlfile != "" and yamlfile is not None:
                    version = open(yamlfile, 'w').write(spec)
                else:  # should really never get here
                    version = open(f"{outdir}/{title}.yaml", 'w').write(spec)
            except IOError:
                Console.error("Unable to write yaml file")
            except Exception as e:
                print(e)

        return


    # TODO: integrate the below functions into the package
    '''
    def file_put(root_url, service, filename, verbose=False):

        url = f'http://{root_url}/cloudmesh/{service}/file/put'
        print("URL", url)
        files = {'file': open(filename, 'rb')}
        r = requests.post(url, files=files)
        return r.text

    def file_list(root_url):
        r = requests.get(f'http://{root_url}/file/list')
        return r.text

    def file_get(root_url, service, filename):

        url = f'http://{root_url}/cloudmesh/{service}/file/get/{filename}'

        print("URL", url)

        r = requests.get(url)
        return r.text
    '''



#!/usr/bin/env python
# ----------------------------------------------------------------------- #
# Copyright 2017, Gregor von Laszewski, Indiana University                #
#                                                                         #
# Licensed under the Apache License, Version 2.0 (the "License"); you may #
# not use this file except in compliance with the License. You may obtain #
# a copy of the License at                                                #
#                                                                         #
# http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                         #
# Unless required by applicable law or agreed to in writing, software     #
# distributed under the License is distributed on an "AS IS" BASIS,       #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.#
# See the License for the specific language governing permissions and     #
# limitations under the License.                                          #
# ------------------------------------------------------------------------#

import io

from setuptools import find_packages, setup


def readfile(filename):
    with io.open(filename, encoding="utf-8") as stream:
        return stream.read().split()


# requiers = readfile ('requirements.txt')
#
# add minimum requirements here
#
requiers = """
connexion[swagger-ui]
cloudmesh-common
cloudmesh-cmd5
cloudmesh-sys
cloudmesh-admin
docstring_parser
numpydoc
""".split("\n")

# dependency_links = ['http://github.com/nicolaiarocci/eve.git@develop']

version = readfile("VERSION")[0].strip()

with open('README.md') as f:
    long_description = f.read()

NAME = "cloudmesh-openapi"
DESCRIPTION = "A command called openapi and foo for the cloudmesh shell"
AUTHOR = "Gregor von Laszewski"
AUTHOR_EMAIL = "laszewski@gmail.com"
URL = "https://github.com/cloudmesh/cloudmesh-openapi"

setup(
        name=NAME,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        description=DESCRIPTION,
        long_description=long_description,
        long_description_content_type="text/markdown",
        version=version,
        license="Apache 2.0",
        url=URL,
        packages=find_packages(),
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Web Environment",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        install_requires=requiers,
        tests_require=[
            "flake8",
            "coverage",
            "tox",
        ],
        zip_safe=False,
        namespace_packages=['cloudmesh'],
    )



###############################################################
# pytest -v --capture=no tests/test_03_generator.py
# pytest -v  tests/test_03_generator.py
# pytest -v --capture=no  tests/test_generator..py::Test_name::<METHODNAME>
###############################################################
"""
# Headline

Here come document for test

"""
import os
import sys

import requests

sys.path.append("./tests")
import util as util
from cloudmesh.common.Shell import Shell
from importlib import import_module
from cloudmesh.openapi.function.executor import Parameter
import types
from cloudmesh.common.dotdict import dotdict
from cloudmesh.common.Benchmark import Benchmark
from cloudmesh.common.util import HEADING




class GeneratorBaseTest:



    def __init__(self,filename,all_functions,import_class,function_name = None):
        global globalcommandstring  # this will hold command string without build folder
        global globalbuildcommandstring # this will hold command string with build folder
        global globalcommandparameter  # this will hold command string without build parameter information
        global globalbuildcommandparameter  # this will hold command string with build parameter information
        globalcommandparameter = Parameter(self.server_dotdict(self.get_servercommand(filename, all_functions, import_class,function_name)))
        build_dotdict=self.build_dotdict(function_name)
        globalbuildcommandparameter=Parameter(build_dotdict)
        globalbuildcommandstring= self.get_build_servercommand(build_dotdict)
        self.service ="openapi"



    def get_servercommand(self,filename,all_functions,import_class,function_name) -> str:
        serverCommand=""
        if import_class:
          serverCommand=f"cms openapi generate --filename={filename} --import_class"
        elif all_functions:
          serverCommand = f"cms openapi generate --filename={filename} --all_functions"
        else:
          serverCommand = f"cms openapi generate {function_name} --filename={filename}"
        return serverCommand

    def get_build_servercommand(self,build_dotdict) -> str:
        serverCommand=""
        if globalbuildcommandparameter.import_class:
          serverCommand=f"cms openapi generate {build_dotdict.FUNCTION} --filename={build_dotdict['--filename']} --import_class"
        elif globalbuildcommandparameter.all_functions:
          serverCommand = f"cms openapi generate --filename={build_dotdict['--filename']} --all_functions"
        else:
          serverCommand = f"cms openapi generate {build_dotdict.FUNCTION} --filename={build_dotdict['--filename']}"
        return serverCommand

    def server_dotdict(self,serverCommand) -> dotdict:
        if serverCommand.__contains__("generate"):
            words=serverCommand.split()
            dotdictd= dotdict()
            for word in words:
                if "--import_class" in word:
                    dotdictd["import_class"] = True
                    dotdictd["all_functions"] = False
                elif "--all_functions" in word:
                    dotdictd["all_functions"] = True
                    dotdictd["import_class"] = False
                elif "--filename" in word:
                    dotdictd["--filename"] = word.split("=")[1]
            return dotdictd
        if serverCommand.__contains__("server"):
           pass

    def build_dotdict(self,function_name = None) -> dotdict:
        dotdictd = dotdict()

        if  globalcommandparameter.import_class:
            dotdictd["import_class"] = True
            dotdictd["all_functions"] = False
            dotdictd["FUNCTION"] = globalcommandparameter.module_name.capitalize()
        elif globalcommandparameter.all_functions:
            dotdictd["all_functions"] = True
            dotdictd["import_class"] = False
        else:
            dotdictd["all_functions"] = False
            dotdictd["import_class"] = False
            dotdictd["FUNCTION"] = function_name

        dotdictd["--filename"] = globalcommandparameter.module_directory+"/build/"+globalcommandparameter.module_name+".py"


        return dotdictd


    def copy_py_file(self):
        import os
        os.makedirs(globalcommandparameter.module_directory+"/build")
        from shutil import copyfile
        HEADING()
        Benchmark.Start()
        copyfile(globalcommandparameter.filename, globalbuildcommandparameter.filename)
        Benchmark.Stop()


    def generate(self):
        """
        function to validate paths information
        """
        HEADING()
        Benchmark.Start()
        Shell.run(globalbuildcommandstring)
        Benchmark.Stop()

    def read_spec(self):
        """
        function to check if YAML synatx is correct or not
        """
        global spec
        HEADING()
        Benchmark.Start()
        spec = util.readyaml(globalbuildcommandparameter.yamlfile)
        keys = spec.keys()
        Benchmark.Stop()
        assert "openapi" in keys
        assert "info" in keys
        assert "servers" in keys
        assert "paths" in keys


    def validate_function(self):
        """
        function to check number of functions are same in py and yaml file.
        """
        HEADING()
        Benchmark.Start()
        sys.path.append(globalbuildcommandparameter.yamldirectory)
        imported_module = import_module(globalbuildcommandparameter.module_name)
        keys = spec.get('paths')
        paths= keys.keys()
        if (globalbuildcommandparameter.all_functions is True):
            for attr_name in dir(imported_module):
                if type(getattr(imported_module, attr_name)).__name__ == 'function':
                    assert f"/{globalbuildcommandparameter.function}/{attr_name}" in paths
        if (globalbuildcommandparameter.import_class is True):
            class_obj = getattr(imported_module, globalbuildcommandparameter.function)
            for attr_name in dir(class_obj):
                attr = getattr(class_obj, attr_name)
                if isinstance(attr, types.MethodType):
                    assert f"/{globalbuildcommandparameter.function}/{attr_name}" in paths

        Benchmark.Stop()

    def delete_file(self):
        HEADING()
        Benchmark.Start()
        import shutil
        shutil.rmtree(globalcommandparameter.module_directory+"/build")
        Benchmark.Stop()


class ServerBaseTest:

    def start_service(self):
        """
        function to test if the server is started and available to return
        a successful http code
        """
        HEADING()
        Benchmark.Start()
        os.system(f"cms openapi server start {globalbuildcommandparameter.yamlfile}")
        result = requests.get('http://127.0.0.1:8080/cloudmesh/ui')
        Benchmark.Stop()
        assert result.status_code == 200  # find test

    def stop_server(self):
        HEADING()

        Benchmark.Start()
        os.system(f"cms openapi server stop {globalbuildcommandparameter.function}")
        Benchmark.Stop()
        gotException = False;
        try:
            result = requests.get('http://127.0.0.1:8080/cloudmesh/ui')
        except Exception as ex:
            gotException = True
        assert gotException


    def benchmark(self):
        Benchmark.print(sysinfo=True, csv=True, tag=service)



import textwrap

from shutil import copyfile

from cloudmesh.common.console import Console
from cloudmesh.common.util import path_expand
from cloudmesh.common.Shell import Shell

# Basic auth
# https://swagger.io/docs/specification/2-0/authentication/basic-authentication/


class BasicAuth:
    """
    This class handles writing the components for basic authentication. 
    This class writes a file to ~/.cloudmesh/.auth_users

    Each row of .auth_users is username:password formatted. 
    """
    HALT_FLAG = '#### basic_auth functionality added'
    USERS_FILE = path_expand('~/.cloudmesh/.auth_users')

    AUTH_FUNC_TEMPLATE = textwrap.dedent("""
    from cloudmesh.openapi.authentication.basic import BasicAuth

    def basic_auth(username, password, required_scopes=None):
        return BasicAuth.basic_auth(username, password)
    
    {halt_flag}
    """)

    @classmethod
    def basic_auth(cls, username, password, required_scopes=None):
        # Scan ~/.cloudmesh/.auth_users for username and password
        users = {}
        with open(cls.USERS_FILE, 'r') as f:
            for row in f:
                user, passwd = row.strip().split(':')
                users[user] = passwd
        if username in users:
            if users[username] == password:
                if username == 'admin':
                    return {'sub': 'admin', 'scope': 'secret'}
                else:
                    return {'sub': 'user1', 'scope': ''}

        return None

    @classmethod
    def reset_users(cls):
        open(cls.USERS_FILE, 'w').close()

    @classmethod
    def add_user(cls, user, password):
        with open(cls.USERS_FILE, 'a') as f:
            f.write(f'{user}:{password}\n')
            f.close()

    @classmethod
    def write_basic_auth(cls, filename, module_name):
        """
        Writes the basic auth configuration to a new python file and returns the new module name
        and new filename
        """
        basic_auth_enabled = False
        for line in open(filename):
            if cls.HALT_FLAG in line:
                basic_auth_enabled = True
                break

        if not basic_auth_enabled:
            filename_auth = filename[:-len('.py')] + '_basic_auth_enabled.py'
            copyfile(filename, filename_auth)
            Console.info(f'copied {filename} to {filename_auth}')
            filename = filename_auth
            module_name = module_name + '_basic_auth_enabled'
            with open(filename, 'a') as f:
                f.write('\n')
                f.write(cls.AUTH_FUNC_TEMPLATE.format(halt_flag=cls.HALT_FLAG))
                f.close()
            Console.info(f'added basic auth functionality to {filename}')

        return module_name, filename



from cloudmesh.common.Printer import Printer
from cloudmesh.common.Shell import Shell
from cloudmesh.mongo.CmDatabase import CmDatabase
from cloudmesh.mongo.DataBaseDecorator import DatabaseUpdate


class Registry:
    """
      This class will help to register service into db.
      which later use to stop server.
    """
    kind = "register"

    collection = "local-registry"

    output = {
        "register": {
            "sort_keys": ["cm.name"],
            "order": ["cm.name",
                      "status",
                      "url",
                      "pid"],
            "header": ["Name",
                       "Status",
                       "Url",
                       "Pid"]
        }
    }

    # noinspection PyPep8Naming
    def Print(self, data, output=None):
        """
        print output in a structured format

        :param data:  input data to be printed out
        :param output:  type of structured output
        :return:  structured output
        """

        if output == "table":

            order = self.output[Registry.kind]['order']  # not pretty
            header = self.output[Registry.kind]['header']  # not pretty
            # humanize = self.output[kind]['humanize']  # not pretty

            print(Printer.flatwrite(data,
                                    sort_keys=["name"],
                                    order=order,
                                    header=header,
                                    output=output,
                                    # humanize=humanize
                                    )
                  )
        else:
            print(Printer.write(data, output=output))

    def __init__(self):
        pass

    @DatabaseUpdate()
    def add(self, name=None, **kwargs):
        """
        add to registry

        :param name: name to be used for registry entry
        :param kwargs:  other optional fields to populate in registry
        :return:  
        """
        entry = {
            "cm": {
                "cloud": "local",
                "kind": "registry",
                "name": name,
                "driver": None
            },
            "name": name,
            "status": "defined"
        }

        for key in kwargs:
            entry[key] = kwargs[key]

        return entry

    def add_form_file(self, filename, **kwargs):
        """
        add to registry from file

        :param filename: file name including path
        :return:  entry to be inserted into Registry
        """

        spec = filename

        title = spec["info"]["title"]

        registry = Registry()
        entry = registry.add(name=title, **kwargs)

        return entry

    def delete(self, name=None):
        """
        delete item from registry

        :param name: name of the item in registry
        :return:  
        """
        cm = CmDatabase()

        collection = cm.collection(self.collection)
        if name is None:
            query = {}
        else:
            query = {'name': name}
        entries = cm.delete(collection=self.collection, **query)
        return entries


    def list(self, name=None):
        """
        list entries in the registry

        :param name:  name of registered server.  If not passed will list all registered servers.
        :return:  list of registered server(s)
        """

        cm = CmDatabase()
        if name == None:
            entries = cm.find(cloud="local", kind="registry")
        else:
            entries = cm.find_name(name=name, kind="registry")

        return entries


    # TODO: determine if these are still needed as these functions are handled by cms already
    '''
    def start(self):
        """
        start the registry

        possibly not needed as we have cms start

        :return:
        """
        r = Shell.cms("start")

    def stop(self):
        """
        stop the registry

        possibly not needed as we have cms start
        this will not just sto the registry but mongo

        :return:
        """
        r = Shell.cms("stop")
    '''
